#!/usr/bin/env python
"""
Runs BFAST on single-end or paired-end data.
TODO: more documentation

TODO:
    - auto-detect gzip or bz2
    - split options (?)
    - queue lengths (?)
    - assumes reference always has been indexed
    - main and secondary indexes
    - scoring matrix file ?
    - read group file ?

usage: bfast_wrapper.py [options]
    -r, --ref=r: The reference genome to use or index
    -f, --fastq=f: The fastq file to use for the mapping
    -F, --output=u: The file to save the output (SAM format)
    -s, --fileSource=s: Whether to use a previously indexed reference sequence or one from history (indexed or history)
    -p, --params=p: Parameter setting to use (pre_set or full)
    -n, --numThreads=n: The number of threads to use
    -A, --space=A: The encoding space (0: base 1: color)
    -o, --offsets=o: The offsets for 'match'
    -l, --loadAllIndexes=l: Load all indexes into memory
    -k, --keySize=k: truncate key size in 'match'
    -K, --maxKeyMatches=K: the maximum number of matches to allow before a key is ignored
    -M, --maxNumMatches=M: the maximum number of matches to allow before the read is discarded
    -w, --whichStrand=w: the strands to consider (0: both 1: forward 2: reverse)
    -t, --timing=t: output timing information to stderr
    -u, --ungapped=u: performed ungapped local alignment
    -U, --unconstrained=U: performed local alignment without mask constraints
    -O, --offset=O: the number of bases before and after each hit to consider in local alignment
    -q, --avgMismatchQuality=q: average mismatch quality
    -a, --algorithm=a: post processing algorithm (0: no filtering, 1: all passing filters, 2: unique, 3: best scoring unique, 4: best score all)
    -P, --disallowPairing=P: do not choose alignments based on pairing
    -R, --reverse=R: paired end reads are given on reverse strands
    -z, --random=z: output a random best scoring alignment
    -D, --dbkey=D: Dbkey for reference genome
    -H, --suppressHeader=H: Suppress the sam header
"""

import optparse
import os
import shutil
import subprocess
import sys
import tempfile


def stop_err(msg):
    sys.stderr.write("%s\n" % msg)
    sys.exit()


def __main__():
    parser = optparse.OptionParser()
    parser.add_option("-r", "--ref", dest="ref", help="The reference genome to index and use")
    parser.add_option("-f", "--fastq", dest="fastq", help="The fastq file to use for the mapping")
    parser.add_option("-F", "--output", dest="output", help="The file to save the output (SAM format)")
    parser.add_option(
        "-A",
        "--space",
        dest="space",
        type="choice",
        default="0",
        choices=("0", "1"),
        help="The encoding space (0: base 1: color)",
    )
    parser.add_option(
        "-H", "--suppressHeader", action="store_true", dest="suppressHeader", default=False, help="Suppress header"
    )
    parser.add_option(
        "-n", "--numThreads", dest="numThreads", type="int", default="1", help="The number of threads to use"
    )
    parser.add_option(
        "-t", "--timing", action="store_true", default=False, dest="timing", help="output timming information to stderr"
    )
    parser.add_option(
        "-l",
        "--loadAllIndexes",
        action="store_true",
        default=False,
        dest="loadAllIndexes",
        help="Load all indexes into memory",
    )
    parser.add_option(
        "-m", "--indexMask", dest="indexMask", help="String containing info on how to build custom indexes"
    )
    parser.add_option(
        "-b",
        "--buildIndex",
        action="store_true",
        dest="buildIndex",
        default=False,
        help="String containing info on how to build custom indexes",
    )
    parser.add_option(
        "--indexRepeatMasker",
        action="store_true",
        dest="indexRepeatMasker",
        default=False,
        help="Do not index lower case sequences. Such as those created by RepeatMasker",
    )
    parser.add_option(
        "--indexContigOptions",
        dest="indexContigOptions",
        default="",
        help="The contig range options to use for the indexing",
    )
    parser.add_option(
        "--indexExonsFileName", dest="indexExonsFileName", default="", help="The exons file to use for the indexing"
    )

    parser.add_option("-o", "--offsets", dest="offsets", default="", help="The offsets for 'match'")
    parser.add_option("-k", "--keySize", dest="keySize", type="int", default="-1", help="truncate key size in 'match'")
    parser.add_option(
        "-K",
        "--maxKeyMatches",
        dest="maxKeyMatches",
        type="int",
        default="-1",
        help="the maximum number of matches to allow before a key is ignored",
    )
    parser.add_option(
        "-M",
        "--maxNumMatches",
        dest="maxNumMatches",
        type="int",
        default="-1",
        help="the maximum number of matches to allow bfore the read is discarded",
    )
    parser.add_option(
        "-w",
        "--whichStrand",
        dest="whichStrand",
        type="choice",
        default="0",
        choices=("0", "1", "2"),
        help="the strands to consider (0: both 1: forward 2: reverse)",
    )

    parser.add_option(
        "--scoringMatrixFileName", dest="scoringMatrixFileName", help="Scoring Matrix file used to score the alignments"
    )
    parser.add_option(
        "-u",
        "--ungapped",
        dest="ungapped",
        action="store_true",
        default=False,
        help="performed ungapped local alignment",
    )
    parser.add_option(
        "-U",
        "--unconstrained",
        dest="unconstrained",
        action="store_true",
        default=False,
        help="performed local alignment without mask constraints",
    )
    parser.add_option(
        "-O",
        "--offset",
        dest="offset",
        type="int",
        default="0",
        help="the number of bases before and after each hit to consider in local alignment",
    )
    parser.add_option(
        "-q",
        "--avgMismatchQuality",
        type="int",
        default="-1",
        dest="avgMismatchQuality",
        help="average mismatch quality",
    )

    parser.add_option(
        "-a",
        "--algorithm",
        dest="algorithm",
        default="0",
        type="choice",
        choices=("0", "1", "2", "3", "4"),
        help="post processing algorithm (0: no filtering, 1: all passing filters, 2: unique, 3: best scoring unique, 4: best score all",
    )
    parser.add_option(
        "--unpaired",
        dest="unpaired",
        action="store_true",
        default=False,
        help="do not choose alignments based on pairing",
    )
    parser.add_option(
        "--reverseStrand",
        dest="reverseStrand",
        action="store_true",
        default=False,
        help="paired end reads are given on reverse strands",
    )
    parser.add_option(
        "--pairedEndInfer",
        dest="pairedEndInfer",
        action="store_true",
        default=False,
        help="break ties when one end of a paired end read by estimating the insert size distribution",
    )
    parser.add_option(
        "--randomBest",
        dest="randomBest",
        action="store_true",
        default=False,
        help="output a random best scoring alignment",
    )

    (options, args) = parser.parse_args()

    # output version # of tool
    try:
        tmp = tempfile.NamedTemporaryFile().name
        tmp_stdout = open(tmp, "wb")
        proc = subprocess.Popen(args="bfast 2>&1", shell=True, stdout=tmp_stdout)
        tmp_stdout.close()
        returncode = proc.wait()
        stdout = None
        for line in open(tmp_stdout.name, "rb"):
            if line.lower().find("version") >= 0:
                stdout = line.strip()
                break
        if stdout:
            sys.stdout.write("%s\n" % stdout)
        else:
            raise Exception
    except Exception:
        sys.stdout.write("Could not determine BFAST version\n")

    buffsize = 1048576

    # make temp directory for bfast, requires trailing slash
    tmp_dir = "%s/" % tempfile.mkdtemp()

    # 'generic' options used in all bfast commands here
    if options.timing:
        all_cmd_options = "-t"
    else:
        all_cmd_options = ""

    try:
        if options.buildIndex:
            reference_filepath = tempfile.NamedTemporaryFile(dir=tmp_dir, suffix=".fa").name
            # build bfast indexes
            os.symlink(options.ref, reference_filepath)

            # bfast fast2brg
            try:
                nuc_space = ["0"]
                if options.space == "1":
                    # color space localalign appears to require nuc space brg
                    nuc_space.append("1")
                for space in nuc_space:
                    cmd = 'bfast fasta2brg -f "%s" -A "%s" %s' % (reference_filepath, space, all_cmd_options)
                    tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                    tmp_stderr = open(tmp, "wb")
                    proc = subprocess.Popen(args=cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                    returncode = proc.wait()
                    tmp_stderr.close()
                    # get stderr, allowing for case where it's very large
                    tmp_stderr = open(tmp, "rb")
                    stderr = ""
                    try:
                        while True:
                            stderr += tmp_stderr.read(buffsize)
                            if not stderr or len(stderr) % buffsize != 0:
                                break
                    except OverflowError:
                        pass
                    tmp_stderr.close()
                    if returncode != 0:
                        raise Exception(stderr)
            except Exception as e:
                raise Exception("Error in 'bfast fasta2brg'.\n" + str(e))

            # bfast index
            try:
                all_index_cmds = 'bfast index %s -f "%s" -A "%s" -n "%s"' % (
                    all_cmd_options,
                    reference_filepath,
                    options.space,
                    options.numThreads,
                )

                if options.indexRepeatMasker:
                    all_index_cmds += " -R"

                if options.indexContigOptions:
                    index_contig_options = [int(_) for _ in options.indexContigOptions.split(",")]
                    if index_contig_options[0] >= 0:
                        all_index_cmds += ' -s "%s"' % index_contig_options[0]
                    if index_contig_options[1] >= 0:
                        all_index_cmds += ' -S "%s"' % index_contig_options[1]
                    if index_contig_options[2] >= 0:
                        all_index_cmds += ' -e "%s"' % index_contig_options[2]
                    if index_contig_options[3] >= 0:
                        all_index_cmds += ' -E "%s"' % index_contig_options[3]
                elif options.indexExonsFileName:
                    all_index_cmds += ' -x "%s"' % options.indexExonsFileName

                index_count = 1
                for mask, hash_width in [mask.split(":") for mask in options.indexMask.split(",")]:
                    cmd = '%s -m "%s" -w "%s" -i "%i"' % (all_index_cmds, mask, hash_width, index_count)
                    tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                    tmp_stderr = open(tmp, "wb")
                    proc = subprocess.Popen(args=cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                    returncode = proc.wait()
                    tmp_stderr.close()
                    # get stderr, allowing for case where it's very large
                    tmp_stderr = open(tmp, "rb")
                    stderr = ""
                    try:
                        while True:
                            stderr += tmp_stderr.read(buffsize)
                            if not stderr or len(stderr) % buffsize != 0:
                                break
                    except OverflowError:
                        pass
                    tmp_stderr.close()
                    if returncode != 0:
                        raise Exception(stderr)
                    index_count += 1
            except Exception as e:
                raise Exception("Error in 'bfast index'.\n" + str(e))

        else:
            reference_filepath = options.ref
        assert reference_filepath and os.path.exists(reference_filepath), "A valid genome reference was not provided."

        # set up aligning and generate aligning command options
        # set up temp output files
        tmp_bmf = tempfile.NamedTemporaryFile(dir=tmp_dir)
        tmp_bmf_name = tmp_bmf.name
        tmp_bmf.close()
        tmp_baf = tempfile.NamedTemporaryFile(dir=tmp_dir)
        tmp_baf_name = tmp_baf.name
        tmp_baf.close()

        bfast_match_cmd = 'bfast match -f "%s" -r "%s" -n "%s" -A "%s" -T "%s" -w "%s" %s' % (
            reference_filepath,
            options.fastq,
            options.numThreads,
            options.space,
            tmp_dir,
            options.whichStrand,
            all_cmd_options,
        )
        bfast_localalign_cmd = 'bfast localalign -f "%s" -m "%s" -n "%s" -A "%s" -o "%s" %s' % (
            reference_filepath,
            tmp_bmf_name,
            options.numThreads,
            options.space,
            options.offset,
            all_cmd_options,
        )
        bfast_postprocess_cmd = 'bfast postprocess -O 1 -f "%s" -i "%s" -n "%s" -A "%s" -a "%s" %s' % (
            reference_filepath,
            tmp_baf_name,
            options.numThreads,
            options.space,
            options.algorithm,
            all_cmd_options,
        )

        if options.offsets:
            bfast_match_cmd += ' -o "%s"' % options.offsets
        if options.keySize >= 0:
            bfast_match_cmd += ' -k "%s"' % options.keySize
        if options.maxKeyMatches >= 0:
            bfast_match_cmd += ' -K "%s"' % options.maxKeyMatches
        if options.maxNumMatches >= 0:
            bfast_match_cmd += ' -M "%s"' % options.maxNumMatches
            bfast_localalign_cmd += ' -M "%s"' % options.maxNumMatches
        if options.scoringMatrixFileName:
            bfast_localalign_cmd += ' -x "%s"' % options.scoringMatrixFileName
            bfast_postprocess_cmd += ' -x "%s"' % options.scoringMatrixFileName
        if options.ungapped:
            bfast_localalign_cmd += " -u"
        if options.unconstrained:
            bfast_localalign_cmd += " -U"
        if options.avgMismatchQuality >= 0:
            bfast_localalign_cmd += ' -q "%s"' % options.avgMismatchQuality
            bfast_postprocess_cmd += ' -q "%s"' % options.avgMismatchQuality
        if options.algorithm == 3:
            if options.pairedEndInfer:
                bfast_postprocess_cmd += " -P"
            if options.randomBest:
                bfast_postprocess_cmd += " -z"
        if options.unpaired:
            bfast_postprocess_cmd += " -U"
        if options.reverseStrand:
            bfast_postprocess_cmd += " -R"

        # instead of using temp files, should we stream through pipes?
        bfast_match_cmd += " > %s" % tmp_bmf_name
        bfast_localalign_cmd += " > %s" % tmp_baf_name
        bfast_postprocess_cmd += " > %s" % options.output

        # need to nest try-except in try-finally to handle 2.4
        try:
            # bfast 'match'
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=bfast_match_cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                raise Exception("Error in 'bfast match'. \n" + str(e))
            # bfast 'localalign'
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=bfast_localalign_cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                raise Exception("Error in 'bfast localalign'. \n" + str(e))
            # bfast 'postprocess'
            try:
                tmp = tempfile.NamedTemporaryFile(dir=tmp_dir).name
                tmp_stderr = open(tmp, "wb")
                proc = subprocess.Popen(args=bfast_postprocess_cmd, shell=True, cwd=tmp_dir, stderr=tmp_stderr.fileno())
                returncode = proc.wait()
                tmp_stderr.close()
                # get stderr, allowing for case where it's very large
                tmp_stderr = open(tmp, "rb")
                stderr = ""
                try:
                    while True:
                        stderr += tmp_stderr.read(buffsize)
                        if not stderr or len(stderr) % buffsize != 0:
                            break
                except OverflowError:
                    pass
                tmp_stderr.close()
                if returncode != 0:
                    raise Exception(stderr)
            except Exception as e:
                raise Exception("Error in 'bfast postprocess'. \n" + str(e))
            # remove header if necessary
            if options.suppressHeader:
                tmp_out = tempfile.NamedTemporaryFile(dir=tmp_dir)
                tmp_out_name = tmp_out.name
                tmp_out.close()
                try:
                    shutil.move(options.output, tmp_out_name)
                except Exception as e:
                    raise Exception("Error moving output file before removing headers. \n" + str(e))
                fout = open(options.output, "w")
                for line in open(tmp_out.name):
                    if len(line) < 3 or line[0:3] not in ["@HD", "@SQ", "@RG", "@PG", "@CO"]:
                        fout.write(line)
                fout.close()
            # check that there are results in the output file
            if os.path.getsize(options.output) > 0:
                if "0" == options.space:
                    sys.stdout.write("BFAST run on Base Space data")
                else:
                    sys.stdout.write("BFAST run on Color Space data")
            else:
                raise Exception(
                    "The output file is empty. You may simply have no matches, or there may be an error with your input file or settings."
                )
        except Exception as e:
            stop_err("The alignment failed.\n" + str(e))
    finally:
        # clean up temp dir
        if os.path.exists(tmp_dir):
            shutil.rmtree(tmp_dir)


if __name__ == "__main__":
    __main__()



import logging
import os
import os.path
from typing import (
    Iterable,
    List,
    Tuple,
    Union,
)

import galaxy.tools.parameters.basic
import galaxy.tools.parameters.grouping
from galaxy.tool_util.verify.interactor import (
    InvalidToolTestDict,
    ToolTestDescription,
    ValidToolTestDict,
)
from galaxy.util import (
    string_as_bool,
    string_as_bool_or_none,
    unicodify,
)

log = logging.getLogger(__name__)


def parse_tests(tool, tests_source) -> Iterable[ToolTestDescription]:
    """
    Build ToolTestDescription objects for each "<test>" elements and
    return default interactor (if any).
    """
    raw_tests_dict = tests_source.parse_tests_to_dict()
    tests: List[ToolTestDescription] = []
    for i, raw_test_dict in enumerate(raw_tests_dict.get("tests", [])):
        test = description_from_tool_object(tool, i, raw_test_dict)
        tests.append(test)
    return tests


def description_from_tool_object(tool, test_index, raw_test_dict) -> ToolTestDescription:
    required_files: List[Tuple[str, dict]] = []
    required_data_tables: List[str] = []
    required_loc_files: List[str] = []

    num_outputs = raw_test_dict.get("expect_num_outputs", None)
    if num_outputs:
        num_outputs = int(num_outputs)
    maxseconds = raw_test_dict.get("maxseconds", None)
    if maxseconds is not None:
        maxseconds = int(maxseconds)

    processed_test_dict: Union[ValidToolTestDict, InvalidToolTestDict]
    try:
        processed_inputs = _process_raw_inputs(
            tool, tool.inputs, raw_test_dict["inputs"], required_files, required_data_tables, required_loc_files
        )

        processed_test_dict = ValidToolTestDict(
            {
                "inputs": processed_inputs,
                "outputs": raw_test_dict["outputs"],
                "output_collections": raw_test_dict["output_collections"],
                "num_outputs": num_outputs,
                "command_line": raw_test_dict.get("command", None),
                "command_version": raw_test_dict.get("command_version", None),
                "stdout": raw_test_dict.get("stdout", None),
                "stderr": raw_test_dict.get("stderr", None),
                "expect_exit_code": raw_test_dict.get("expect_exit_code", None),
                "expect_failure": raw_test_dict.get("expect_failure", False),
                "expect_test_failure": raw_test_dict.get("expect_test_failure", False),
                "required_files": required_files,
                "required_data_tables": required_data_tables,
                "required_loc_files": required_loc_files,
                "tool_id": tool.id,
                "tool_version": tool.version,
                "test_index": test_index,
                "maxseconds": maxseconds,
                "error": False,
            }
        )
    except Exception as e:
        log.exception("Failed to load tool test number [%d] for %s" % (test_index, tool.id))
        processed_test_dict = InvalidToolTestDict(
            {
                "tool_id": tool.id,
                "tool_version": tool.version,
                "test_index": test_index,
                "inputs": {},
                "error": True,
                "exception": unicodify(e),
                "maxseconds": maxseconds,
            }
        )

    return ToolTestDescription(processed_test_dict)


def _process_raw_inputs(
    tool, tool_inputs, raw_inputs, required_files, required_data_tables, required_loc_files, parent_context=None
):
    """
    Recursively expand flat list of inputs into "tree" form of flat list
    (| using to nest to new levels) structure and expand dataset
    information as proceeding to populate self.required_files.
    """
    parent_context = parent_context or RootParamContext()
    expanded_inputs = {}
    for value in tool_inputs.values():
        if isinstance(value, galaxy.tools.parameters.grouping.Conditional):
            cond_context = ParamContext(name=value.name, parent_context=parent_context)
            assert value.test_param
            case_context = ParamContext(name=value.test_param.name, parent_context=cond_context)
            raw_input_dict = case_context.extract_value(raw_inputs)
            case_value = raw_input_dict["value"] if raw_input_dict else None
            case = _matching_case_for_value(tool, value, case_value)
            if case:
                for input_name, input_value in case.inputs.items():
                    case_inputs = _process_raw_inputs(
                        tool,
                        {input_name: input_value},
                        raw_inputs,
                        required_files,
                        required_data_tables,
                        required_loc_files,
                        parent_context=cond_context,
                    )
                    expanded_inputs.update(case_inputs)
                if not value.type == "text":
                    expanded_case_value = _split_if_str(case.value)
                if case_value is not None:
                    # A bit tricky here - we are growing inputs with value
                    # that may be implicit (i.e. not defined by user just
                    # a default defined in tool). So we do not want to grow
                    # expanded_inputs and risk repeat block viewing this
                    # as a new instance with value defined and hence enter
                    # an infinite loop - hence the "case_value is not None"
                    # check.
                    processed_value = _process_simple_value(
                        value.test_param, expanded_case_value, required_data_tables, required_loc_files
                    )
                    expanded_inputs[case_context.for_state()] = processed_value
        elif isinstance(value, galaxy.tools.parameters.grouping.Section):
            context = ParamContext(name=value.name, parent_context=parent_context)
            assert value.inputs
            for r_value in value.inputs.values():
                expanded_input = _process_raw_inputs(
                    tool,
                    {context.for_state(): r_value},
                    raw_inputs,
                    required_files,
                    required_data_tables,
                    required_loc_files,
                    parent_context=context,
                )
                if expanded_input:
                    expanded_inputs.update(expanded_input)
        elif isinstance(value, galaxy.tools.parameters.grouping.Repeat):
            repeat_index = 0
            while True:
                context = ParamContext(name=value.name, index=repeat_index, parent_context=parent_context)
                updated = False
                assert value.inputs
                for r_value in value.inputs.values():
                    expanded_input = _process_raw_inputs(
                        tool,
                        {context.for_state(): r_value},
                        raw_inputs,
                        required_files,
                        required_data_tables,
                        required_loc_files,
                        parent_context=context,
                    )
                    if expanded_input:
                        expanded_inputs.update(expanded_input)
                        updated = True
                if not updated:
                    break
                repeat_index += 1
        else:
            context = ParamContext(name=value.name, parent_context=parent_context)
            raw_input_dict = context.extract_value(raw_inputs)
            if raw_input_dict:
                name = raw_input_dict["name"]
                param_value = raw_input_dict["value"]
                param_extra = raw_input_dict["attributes"]
                location = param_extra.get("location")
                if not value.type == "text":
                    param_value = _split_if_str(param_value)
                if isinstance(value, galaxy.tools.parameters.basic.DataToolParameter):
                    if location and value.multiple:
                        # We get the input/s from the location which can be a list of urls separated by commas
                        locations = _split_if_str(location)
                        param_value = []
                        for location in locations:
                            v = os.path.basename(location)
                            param_value.append(v)
                            # param_extra should contain only the corresponding location
                            extra = dict(param_extra)
                            extra["location"] = location
                            _add_uploaded_dataset(context.for_state(), v, extra, value, required_files)
                    else:
                        if not isinstance(param_value, list):
                            param_value = [param_value]
                        for v in param_value:
                            _add_uploaded_dataset(context.for_state(), v, param_extra, value, required_files)
                    processed_value = param_value
                elif isinstance(value, galaxy.tools.parameters.basic.DataCollectionToolParameter):
                    assert "collection" in param_extra
                    collection_def = param_extra["collection"]
                    for input_dict in collection_def.collect_inputs():
                        name = input_dict["name"]
                        value = input_dict["value"]
                        attributes = input_dict["attributes"]
                        require_file(name, value, attributes, required_files)
                    processed_value = collection_def
                else:
                    processed_value = _process_simple_value(
                        value, param_value, required_data_tables, required_loc_files
                    )
                expanded_inputs[context.for_state()] = processed_value
    return expanded_inputs


def _process_simple_value(param, param_value, required_data_tables, required_loc_files):
    if isinstance(param, galaxy.tools.parameters.basic.SelectToolParameter):
        # Tests may specify values as either raw value or the value
        # as they appear in the list - the API doesn't and shouldn't
        # accept the text value - so we need to convert the text
        # into the form value.
        def process_param_value(param_value):
            found_value = False
            value_for_text = None
            for text, opt_value, _ in getattr(param, "static_options", []):
                if param_value == opt_value:
                    found_value = True
                if value_for_text is None and param_value == text:
                    value_for_text = opt_value
            if param.options and not isinstance(param, galaxy.tools.parameters.basic.DrillDownSelectToolParameter):
                if param.options.tool_data_table_name:
                    required_data_tables.append(param.options.tool_data_table_name)
                elif param.options.index_file:
                    required_loc_files.append(param.options.index_file)
            if not found_value and value_for_text is not None:
                processed_value = value_for_text
            else:
                processed_value = param_value
            return processed_value

        # Do replacement described above for lists or singleton
        # values.
        if isinstance(param_value, list):
            processed_value = list(map(process_param_value, param_value))
        else:
            processed_value = process_param_value(param_value)
    elif isinstance(param, galaxy.tools.parameters.basic.BooleanToolParameter):
        # Like above, tests may use the tool define values of simply
        # true/false.
        processed_value = _process_bool_param_value(param, param_value)
    else:
        processed_value = param_value
    return processed_value


def _matching_case_for_value(tool, cond, declared_value):
    test_param = cond.test_param
    if isinstance(test_param, galaxy.tools.parameters.basic.BooleanToolParameter):
        if declared_value is None:
            # No explicit value for param in test case, determine from default
            query_value = test_param.checked
        else:
            query_value = _process_bool_param_value(test_param, declared_value)

        def matches_declared_value(case_value):
            return _process_bool_param_value(test_param, case_value) == query_value

    elif isinstance(test_param, galaxy.tools.parameters.basic.SelectToolParameter):
        if declared_value is not None:
            # Test case supplied explicit value to check against.

            def matches_declared_value(case_value):
                return case_value == declared_value

        elif test_param.static_options:
            # No explicit value in test case, not much to do if options are dynamic but
            # if static options are available can find the one specified as default or
            # fallback on top most option (like GUI).
            for name, _, selected in test_param.static_options:
                if selected:
                    default_option = name
            else:
                first_option = test_param.static_options[0]
                first_option_value = first_option[1]
                default_option = first_option_value

            def matches_declared_value(case_value):
                return case_value == default_option

        else:
            # No explicit value for this param and cannot determine a
            # default - give up. Previously this would just result in a key
            # error exception.
            msg = f"Failed to find test parameter value specification required for conditional {cond.name}"
            raise Exception(msg)

    # Check the tool's defined cases against predicate to determine
    # selected or default.
    for case in cond.cases:
        if matches_declared_value(case.value):
            return case
    else:
        msg_template = "%s - Failed to find case matching value (%s) for test parameter specification for conditional %s. Remainder of test behavior is unspecified."
        msg = msg_template % (tool.id, declared_value, cond.name)
        log.info(msg)


def _add_uploaded_dataset(name, value, extra, input_parameter, required_files):
    if value is None:
        assert input_parameter.optional, f"{name} is not optional. You must provide a valid filename."
        return value
    return require_file(name, value, extra, required_files)


def _split_if_str(value):
    split = isinstance(value, str)
    if split:
        value = value.split(",")
    return value


def _process_bool_param_value(param, param_value):
    assert isinstance(param, galaxy.tools.parameters.basic.BooleanToolParameter)
    was_list = False
    if isinstance(param_value, list):
        was_list = True
        param_value = param_value[0]
    if param.truevalue == param_value:
        processed_value = True
    elif param.falsevalue == param_value:
        processed_value = False
    else:
        if param.optional:
            processed_value = string_as_bool_or_none(param_value)
        else:
            processed_value = string_as_bool(param_value)
    return [processed_value] if was_list else processed_value


def require_file(name, value, extra, required_files):
    if (value, extra) not in required_files:
        required_files.append((value, extra))  # these files will be uploaded
    name_change = [att for att in extra.get("edit_attributes", []) if att.get("type") == "name"]
    if name_change:
        name_change = name_change[-1].get("value")  # only the last name change really matters
        value = name_change  # change value for select to renamed uploaded file for e.g. composite dataset
    else:
        for end in [".zip", ".gz"]:
            if value.endswith(end):
                value = value[: -len(end)]
                break
        value = os.path.basename(value)  # if uploading a file in a path other than root of test-data
    return value


class ParamContext:
    def __init__(self, name, index=None, parent_context=None):
        self.parent_context = parent_context
        self.name = name
        self.index = None if index is None else int(index)

    def for_state(self):
        name = self.name if self.index is None else "%s_%d" % (self.name, self.index)
        if parent_for_state := self.parent_context.for_state():
            return f"{parent_for_state}|{name}"
        else:
            return name

    def __str__(self):
        return f"Context[for_state={self.for_state()}]"

    def param_names(self):
        for parent_context_param in self.parent_context.param_names():
            if self.index is not None:
                yield "%s|%s_%d" % (parent_context_param, self.name, self.index)
            else:
                yield f"{parent_context_param}|{self.name}"
        if self.index is not None:
            yield "%s_%d" % (self.name, self.index)
        else:
            yield self.name

    def extract_value(self, raw_inputs):
        for param_name in self.param_names():
            value = self.__raw_param_found(param_name, raw_inputs)
            if value:
                return value
        return None

    def __raw_param_found(self, param_name, raw_inputs):
        index = None
        for i, raw_input_dict in enumerate(raw_inputs):
            if raw_input_dict["name"] == param_name:
                index = i
        if index is not None:
            raw_input_dict = raw_inputs[index]
            del raw_inputs[index]
            return raw_input_dict
        else:
            return None


class RootParamContext:
    def __init__(self):
        pass

    def for_state(self):
        return ""

    def param_names(self):
        return []

    def get_index(self):
        return 0



"""Utilities defining "Galaxy Flavored Markdown".

This is an extension of markdown designed to allow rendering Galaxy object
references.

The core "Galaxy Flavored Markdown" format should just reference objects
by encoded IDs - but preprocessing should allow for instance workflow objects
to be referenced relative to the workflow (inputs, outputs, steps, etc..) and
potential history flavor would allow objects to be referenced by HID. This
second idea is unimplemented, it is just an example of the general concept of
context specific processing.
"""

import abc
import base64
import codecs
import logging
import os
import re
import shutil
import tempfile
from typing import (
    Any,
    Dict,
    List,
    Match,
    Optional,
)

import markdown

try:
    import weasyprint
except Exception:
    weasyprint = None

from galaxy.config import GalaxyAppConfiguration
from galaxy.exceptions import (
    MalformedContents,
    ObjectNotFound,
    ServerNotConfiguredForRequest,
)
from galaxy.managers.hdcas import HDCASerializer
from galaxy.managers.jobs import (
    JobManager,
    summarize_job_metrics,
    summarize_job_parameters,
)
from galaxy.managers.licenses import LicensesManager
from galaxy.model import Job
from galaxy.model.item_attrs import get_item_annotation_str
from galaxy.model.orm.now import now
from galaxy.schema import PdfDocumentType
from galaxy.schema.tasks import GeneratePdfDownload
from galaxy.short_term_storage import (
    ShortTermStorageMonitor,
    storage_context,
)
from galaxy.util.markdown import literal_via_fence
from galaxy.util.resources import resource_string
from galaxy.util.sanitize_html import sanitize_html
from .markdown_parse import (
    GALAXY_MARKDOWN_FUNCTION_CALL_LINE,
    validate_galaxy_markdown,
)

log = logging.getLogger(__name__)

ARG_VAL_CAPTURED_REGEX = r"""(?:([\w_\-\|]+)|\"([^\"]+)\"|\'([^\']+)\')"""
OUTPUT_LABEL_PATTERN = re.compile(rf"output=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
INPUT_LABEL_PATTERN = re.compile(rf"input=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
STEP_LABEL_PATTERN = re.compile(rf"step=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
PATH_LABEL_PATTERN = re.compile(rf"path=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
SIZE_PATTERN = re.compile(rf"size=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
# STEP_OUTPUT_LABEL_PATTERN = re.compile(r'step_output=([\w_\-]+)/([\w_\-]+)')
UNENCODED_ID_PATTERN = re.compile(
    r"(history_id|workflow_id|history_dataset_id|history_dataset_collection_id|job_id|implicit_collection_jobs_id|invocation_id)=([\d]+)"
)
ENCODED_ID_PATTERN = re.compile(
    r"(history_id|workflow_id|history_dataset_id|history_dataset_collection_id|job_id|implicit_collection_jobs_id|invocation_id)=([a-z0-9]+)"
)
INVOCATION_SECTION_MARKDOWN_CONTAINER_LINE_PATTERN = re.compile(r"```\s*galaxy\s*")
GALAXY_FENCED_BLOCK = re.compile(r"^```\s*galaxy\s*(.*?)^```", re.MULTILINE ^ re.DOTALL)
VALID_CONTAINER_START_PATTERN = re.compile(r"^```\s+[\w]+.*$")


def ready_galaxy_markdown_for_import(trans, external_galaxy_markdown):
    """Convert from encoded IDs to decoded numeric IDs for storing in the DB."""

    _validate(external_galaxy_markdown, internal=False)

    def _remap(container, line):
        object_id = None
        if id_match := re.search(ENCODED_ID_PATTERN, line):
            object_id = id_match.group(2)
            decoded_id = trans.security.decode_id(object_id)
            line = line.replace(id_match.group(), "%s=%d" % (id_match.group(1), decoded_id))
        return (line, False)

    internal_markdown = _remap_galaxy_markdown_calls(_remap, external_galaxy_markdown)
    return internal_markdown


class GalaxyInternalMarkdownDirectiveHandler(metaclass=abc.ABCMeta):
    def walk(self, trans, internal_galaxy_markdown):
        hda_manager = trans.app.hda_manager
        history_manager = trans.app.history_manager
        workflow_manager = trans.app.workflow_manager
        job_manager = JobManager(trans.app)
        collection_manager = trans.app.dataset_collection_manager

        def _check_object(object_id, line):
            if object_id is None:
                raise MalformedContents(f"Missing object identifier [{line}].")

        def _remap(container, line):
            line, object_id, encoded_id = self._encode_line(trans, line)
            if container == "history_link":
                _check_object(object_id, line)
                history = history_manager.get_accessible(object_id, trans.user)
                rval = self.handle_history_link(line, history)
            elif container == "history_dataset_display":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_display(line, hda)
            elif container == "history_dataset_link":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_display(line, hda)
            elif container == "history_dataset_index":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_display(line, hda)
            elif container == "history_dataset_embedded":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_embedded(line, hda)
            elif container == "history_dataset_as_image":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_as_image(line, hda)
            elif container == "history_dataset_as_table":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_as_table(line, hda)
            elif container == "history_dataset_peek":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_peek(line, hda)
            elif container == "history_dataset_info":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_info(line, hda)
            elif container == "history_dataset_type":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_type(line, hda)
            elif container == "history_dataset_name":
                _check_object(object_id, line)
                hda = hda_manager.get_accessible(object_id, trans.user)
                rval = self.handle_dataset_name(line, hda)
            elif container == "workflow_display":
                stored_workflow = workflow_manager.get_stored_accessible_workflow(trans, encoded_id)
                workflow_version_str = _parse_directive_argument_value("workflow_checkpoint", line)
                workflow_version = None if not workflow_version_str else int(workflow_version_str)
                rval = self.handle_workflow_display(line, stored_workflow, workflow_version)
            elif container == "workflow_image":
                stored_workflow = workflow_manager.get_stored_accessible_workflow(trans, encoded_id)
                workflow_version_str = _parse_directive_argument_value("workflow_checkpoint", line)
                workflow_version = None if not workflow_version_str else int(workflow_version_str)
                rval = self.handle_workflow_image(line, stored_workflow, workflow_version)
            elif container == "workflow_license":
                stored_workflow = workflow_manager.get_stored_accessible_workflow(trans, encoded_id)
                rval = self.handle_workflow_license(line, stored_workflow)
            elif container == "history_dataset_collection_display":
                hdca = collection_manager.get_dataset_collection_instance(trans, "history", encoded_id)
                rval = self.handle_dataset_collection_display(line, hdca)
            elif container == "tool_stdout":
                job = job_manager.get_accessible_job(trans, object_id)
                rval = self.handle_tool_stdout(line, job)
            elif container == "tool_stderr":
                job = job_manager.get_accessible_job(trans, object_id)
                rval = self.handle_tool_stderr(line, job)
            elif container == "job_parameters":
                job = job_manager.get_accessible_job(trans, object_id)
                rval = self.handle_job_parameters(line, job)
            elif container == "job_metrics":
                job = job_manager.get_accessible_job(trans, object_id)
                rval = self.handle_job_metrics(line, job)
            elif container == "generate_galaxy_version":
                version = trans.app.config.version_major
                rval = self.handle_generate_galaxy_version(line, version)
            elif container == "generate_time":
                rval = self.handle_generate_time(line, now())
            elif container == "instance_access_link":
                url = trans.app.config.instance_access_url
                rval = self.handle_instance_access_link(line, url)
            elif container == "instance_resources_link":
                url = trans.app.config.instance_resource_url
                rval = self.handle_instance_resources_link(line, url)
            elif container == "instance_help_link":
                url = trans.app.config.helpsite_url
                rval = self.handle_instance_help_link(line, url)
            elif container == "instance_support_link":
                url = trans.app.config.support_url
                rval = self.handle_instance_support_link(line, url)
            elif container == "instance_citation_link":
                url = trans.app.config.citation_url
                rval = self.handle_instance_citation_link(line, url)
            elif container == "instance_terms_link":
                url = trans.app.config.terms_url
                rval = self.handle_instance_terms_link(line, url)
            elif container == "instance_organization_link":
                title = trans.app.config.organization_name
                url = trans.app.config.organization_url
                rval = self.handle_instance_organization_link(line, title, url)
            elif container == "invocation_time":
                invocation = workflow_manager.get_invocation(trans, object_id)
                rval = self.handle_invocation_time(line, invocation)
            elif container == "visualization":
                rval = None
            else:
                raise MalformedContents(f"Unknown Galaxy Markdown directive encountered [{container}].")
            if rval is not None:
                return rval
            else:
                return (line, False)

        def _remap_container(container, line):
            try:
                return _remap(container, line)
            except Exception as e:
                line, *_ = self._encode_line(trans, line)
                return self.handle_error(container, line, str(e))

        export_markdown = _remap_galaxy_markdown_calls(_remap_container, internal_galaxy_markdown)
        return export_markdown

    def _encode_line(self, trans, line):
        object_id = None
        encoded_id = None
        if id_match := re.search(UNENCODED_ID_PATTERN, line):
            object_id = int(id_match.group(2))
            encoded_id = trans.security.encode_id(object_id)
            line = line.replace(id_match.group(), f"{id_match.group(1)}={encoded_id}")
        return line, object_id, encoded_id

    @abc.abstractmethod
    def handle_history_link(self, line, history):
        pass

    @abc.abstractmethod
    def handle_dataset_display(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_as_image(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_as_table(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_peek(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_embedded(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_info(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_name(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_dataset_type(self, line, hda):
        pass

    @abc.abstractmethod
    def handle_workflow_display(self, line, stored_workflow, workflow_version: Optional[int]):
        pass

    @abc.abstractmethod
    def handle_workflow_image(self, line, stored_workflow, workflow_version: Optional[int]):
        pass

    @abc.abstractmethod
    def handle_workflow_license(self, line, stored_workflow):
        pass

    @abc.abstractmethod
    def handle_dataset_collection_display(self, line, hdca):
        pass

    @abc.abstractmethod
    def handle_tool_stdout(self, line, job):
        pass

    @abc.abstractmethod
    def handle_tool_stderr(self, line, job):
        pass

    @abc.abstractmethod
    def handle_job_metrics(self, line, job):
        pass

    @abc.abstractmethod
    def handle_job_parameters(self, line, job):
        pass

    @abc.abstractmethod
    def handle_generate_galaxy_version(self, line, galaxy_version):
        pass

    @abc.abstractmethod
    def handle_generate_time(self, line, date):
        pass

    @abc.abstractmethod
    def handle_instance_access_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_resources_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_help_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_support_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_citation_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_terms_link(self, line, url):
        pass

    @abc.abstractmethod
    def handle_instance_organization_link(self, line, title, url):
        pass

    @abc.abstractmethod
    def handle_invocation_time(self, line, date):
        pass

    @abc.abstractmethod
    def handle_error(self, container, line, error):
        pass


class ReadyForExportMarkdownDirectiveHandler(GalaxyInternalMarkdownDirectiveHandler):
    def __init__(self, trans, extra_rendering_data=None):
        extra_rendering_data = extra_rendering_data or {}
        self.trans = trans
        self.extra_rendering_data = extra_rendering_data

    def ensure_rendering_data_for(self, object_type, obj):
        encoded_id = self.trans.security.encode_id(obj.id)
        if object_type not in self.extra_rendering_data:
            self.extra_rendering_data[object_type] = {}
        object_type_data = self.extra_rendering_data[object_type]
        if encoded_id not in object_type_data:
            object_type_data[encoded_id] = {}
        return object_type_data[encoded_id]

    def extend_history_dataset_rendering_data(self, obj, key, val, default_val):
        self.ensure_rendering_data_for("history_datasets", obj)[key] = val or default_val

    def handle_dataset_display(self, line, hda):
        self.handle_dataset_name(line, hda)
        self.handle_dataset_type(line, hda)

    def handle_dataset_embedded(self, line, hda):
        self.handle_dataset_name(line, hda)

    def handle_dataset_peek(self, line, hda):
        self.extend_history_dataset_rendering_data(hda, "peek", hda.peek, "*No Dataset Peek Available*")

    def handle_dataset_info(self, line, hda):
        self.extend_history_dataset_rendering_data(hda, "info", hda.info, "*No Dataset Info Available*")

    def handle_workflow_display(self, line, stored_workflow, workflow_version: Optional[int]):
        self.ensure_rendering_data_for("workflows", stored_workflow)["name"] = stored_workflow.name

    def handle_workflow_image(self, line, stored_workflow, workflow_version: Optional[int]):
        pass

    def handle_workflow_license(self, line, stored_workflow):
        self.ensure_rendering_data_for("workflows", stored_workflow)[
            "license"
        ] = stored_workflow.latest_workflow.license

    def handle_dataset_collection_display(self, line, hdca):
        hdca_serializer = HDCASerializer(self.trans.app)
        hdca_view = hdca_serializer.serialize_to_view(hdca, user=self.trans.user, trans=self.trans, view="summary")
        self.ensure_rendering_data_for("history_dataset_collections", hdca).update(hdca_view)

    def handle_tool_stdout(self, line, job):
        self.ensure_rendering_data_for("jobs", job)["tool_stdout"] = job.tool_stdout or "*No Standard Output Available*"

    def handle_tool_stderr(self, line, job):
        self.ensure_rendering_data_for("jobs", job)["tool_stderr"] = job.tool_stderr or "*No Standard Error Available*"

    def handle_history_link(self, line, history):
        self.ensure_rendering_data_for("histories", history)["name"] = history.name

    # Following three cases - the client side widgets have everything they need
    # from the encoded ID. Don't implement a default on the base class though because
    # it is good to force both Client and PDF/HTML export to deal with each new directive
    # explicitly.
    def handle_dataset_as_image(self, line, hda):
        pass

    def handle_dataset_as_table(self, line, hda):
        pass

    def handle_job_metrics(self, line, job):
        pass

    def handle_job_parameters(self, line, job):
        pass

    def handle_generate_galaxy_version(self, line, generate_version):
        pass

    def handle_generate_time(self, line, generate_time):
        pass

    def handle_instance_access_link(self, line, url):
        pass

    def handle_instance_resources_link(self, line, url):
        pass

    def handle_instance_help_link(self, line, url):
        pass

    def handle_instance_support_link(self, line, url):
        pass

    def handle_instance_citation_link(self, line, url):
        pass

    def handle_instance_terms_link(self, line, url):
        pass

    def handle_instance_organization_link(self, line, title, url):
        pass

    def handle_invocation_time(self, line, invocation):
        self.ensure_rendering_data_for("invocations", invocation)["create_time"] = invocation.create_time.strftime(
            "%Y-%m-%d, %H:%M:%S"
        )

    def handle_dataset_type(self, line, hda):
        self.extend_history_dataset_rendering_data(hda, "ext", hda.ext, "*Unknown dataset type*")

    def handle_dataset_name(self, line, hda):
        self.extend_history_dataset_rendering_data(hda, "name", hda.name, "*Unknown dataset name*")

    def handle_error(self, container, line, error):
        if "errors" not in self.extra_rendering_data:
            self.extra_rendering_data["errors"] = []
        self.extra_rendering_data["errors"].append(
            {
                "error": error,
                "line": line,
                "container": container,
            }
        )
        return (line, False)


def ready_galaxy_markdown_for_export(trans, internal_galaxy_markdown):
    """Fill in details needed to render Galaxy flavored markdown.

    Take it from a minimal internal version to an externally render-able version
    with more details populated and actual IDs replaced with encoded IDs to render
    external links. Return expanded markdown and extra data useful for rendering
    custom container tags.
    """
    extra_rendering_data = {
        "generate_time": now().isoformat(),
        "generate_version": trans.app.config.version_major,
    }
    # Walk Galaxy directives inside the Galaxy Markdown and collect dict-ified data
    # needed to render this efficiently.
    directive_handler = ReadyForExportMarkdownDirectiveHandler(trans, extra_rendering_data)
    export_markdown = directive_handler.walk(trans, internal_galaxy_markdown)
    return export_markdown, extra_rendering_data


class ToBasicMarkdownDirectiveHandler(GalaxyInternalMarkdownDirectiveHandler):
    def __init__(self, trans):
        self.trans = trans

    def handle_dataset_display(self, line, hda):
        name = hda.name or ""
        markdown = "---\n"
        markdown += f"**Dataset:** {name}\n\n"
        markdown += self._display_dataset_content(hda)
        markdown += "\n---\n"
        return (markdown, True)

    def handle_dataset_embedded(self, line, hda):
        datatype = hda.datatype
        markdown = ""
        # subtly different than below since no Contents: prefix and new lines and such.
        if datatype is None:
            markdown += "*cannot display - cannot format unknown datatype*\n\n"
        else:
            markdown += datatype.display_as_markdown(hda)
        return (markdown, True)

    def _display_dataset_content(self, hda, header="Contents"):
        datatype = hda.datatype
        markdown = ""
        if datatype is None:
            markdown += f"**{header}:** *cannot display - cannot format unknown datatype*\n\n"
        else:
            markdown += f"**{header}:**\n"
            markdown += datatype.display_as_markdown(hda)
        return markdown

    def handle_dataset_as_image(self, line, hda):
        dataset = hda.dataset
        name = hda.name or ""

        if path_match := re.search(PATH_LABEL_PATTERN, line):
            filepath = path_match.group(2)
            file = os.path.join(hda.extra_files_path, filepath)
        else:
            file = dataset.get_file_name()

        with open(file, "rb") as f:
            image_data = f.read()
        rval = (self._embed_image(name, "png", image_data), True)
        return rval

    def _embed_image(self, name: str, image_type: str, image_data: bytes):
        base64_image_data = base64.b64encode(image_data).decode("utf-8")
        return f"![{name}](data:image/{image_type};base64,{base64_image_data})"

    def handle_dataset_as_table(self, line, hda):
        # TODO: this form of the rendering doesn't do anything special with advanced
        # options yet but could easily be modified in the future. show_column_headers,
        # compact, title, and footer should be handled in here to bring the PDF and the
        # web rendering closer.
        rval = self.handle_dataset_embedded(line, hda)
        return rval

    def handle_history_link(self, line, history):
        if history:
            content = literal_via_fence(history.name)
        else:
            content = "*No History available*"
        return (content, True)

    def handle_dataset_peek(self, line, hda):
        if hda.peek:
            content = literal_via_fence(hda.peek)
        else:
            content = "*No Dataset Peek Available*"
        return (content, True)

    def handle_dataset_info(self, line, hda):
        if hda.info:
            content = literal_via_fence(hda.info)
        else:
            content = "*No Dataset Info Available*"
        return (content, True)

    def handle_workflow_display(self, line, stored_workflow, workflow_version: Optional[int]):
        # simple markdown
        markdown = "---\n"
        markdown += f"**Workflow:** {stored_workflow.name}\n\n"
        markdown += "**Steps:**\n\n"
        markdown += "|Step|Annotation|\n"
        markdown += "|----|----------|\n"
        # Pass two should add tool information, labels, etc.. but
        # it requires module_injector and such.
        workflow = stored_workflow.get_internal_version(workflow_version)
        for order_index, step in enumerate(workflow.steps):
            annotation = get_item_annotation_str(self.trans.sa_session, self.trans.user, step) or ""
            markdown += "|{}|{}|\n".format(step.label or "Step %d" % (order_index + 1), annotation)
        markdown += "\n---\n"
        return (markdown, True)

    def handle_workflow_license(self, line, stored_workflow):
        # workflow_manager = self.trans.app.workflow_manager
        license_manager = LicensesManager()
        markdown = "*No license specified.*"
        if license_id := stored_workflow.latest_workflow.license:
            try:
                license_metadata = license_manager.get_license_by_id(license_id)
                markdown = f"[{license_metadata.name}]({license_metadata.url})"
            except ObjectNotFound:
                markdown = f"Unknown license ({license_id})"
        return (f"\n\n{markdown}\n\n", True)

    def handle_workflow_image(self, line, stored_workflow, workflow_version: Optional[int]):
        workflow_manager = self.trans.app.workflow_manager
        workflow = stored_workflow.get_internal_version(workflow_version)
        image_data = workflow_manager.get_workflow_svg(self.trans, workflow, for_embed=True)
        rval = (self._embed_image("Workflow", "svg+xml", image_data), True)
        return rval

    def handle_dataset_collection_display(self, line, hdca):
        name = hdca.name or ""
        # put it in a list to hack around no nonlocal on Python 2.
        markdown_wrapper = [f"**Dataset Collection:** {name}\n\n"]

        def walk_elements(collection, element_prefix=""):
            if ":" in collection.collection_type:
                for element in collection.elements:
                    walk_elements(element.child_collection, f"{element_prefix + element.element_identifier}:")
            else:
                for element in collection.elements:
                    markdown_wrapper[0] += f"**Element:** {element_prefix}{element.element_identifier}\n\n"
                    markdown_wrapper[0] += self._display_dataset_content(element.hda, header="Element Contents")

        walk_elements(hdca.collection)
        markdown = f"---\n{markdown_wrapper[0]}\n---\n"
        return (markdown, True)

    def handle_tool_stdout(self, line, job):
        stdout = job.tool_stdout or "*No Standard Output Available*"
        return (f"**Standard Output:** {stdout}", True)

    def handle_tool_stderr(self, line, job):
        stderr = job.tool_stderr or "*No Standard Error Available*"
        return (f"**Standard Error:** {stderr}", True)

    def handle_job_metrics(self, line, job):
        job_metrics = summarize_job_metrics(self.trans, job)
        metrics_by_plugin: Dict[str, Dict[str, Any]] = {}
        for job_metric in job_metrics:
            plugin = job_metric["plugin"]
            if plugin not in metrics_by_plugin:
                metrics_by_plugin[plugin] = {}
            metrics_by_plugin[plugin][job_metric["title"]] = job_metric["value"]
        markdown = ""
        for metric_plugin, metrics_for_plugin in metrics_by_plugin.items():
            markdown += f"**{metric_plugin}**\n\n"
            markdown += "|   |   |\n|---|--|\n"
            for title, value in metrics_for_plugin.items():
                markdown += f"| {title} | {value} |\n"
        return (markdown, True)

    def handle_job_parameters(self, line, job: Job):
        markdown = """
| Input Parameter | Value |
|-----------------|-------|
"""
        parameters = summarize_job_parameters(self.trans, job)["parameters"]
        for parameter in parameters:
            markdown += "| "
            depth = parameter["depth"]
            if depth > 1:
                markdown += f"{'>' * (parameter['depth'] - 1)} "
            markdown += parameter["text"]
            markdown += " | "
            value = parameter["value"]
            if isinstance(value, list):
                markdown += ", ".join(f"{p['hid']}: {p['name']}" for p in value)
            else:
                markdown += value
            markdown += " |\n"

        return (markdown, True)

    def handle_generate_galaxy_version(self, line, generate_version):
        if generate_version:
            content = literal_via_fence(generate_version)
        else:
            content = "*No Galaxy Version Available*"
        return (content, True)

    def handle_generate_time(self, line, generate_time):
        content = literal_via_fence(generate_time.isoformat())
        return (content, True)

    def handle_instance_access_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_resources_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_help_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_support_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_citation_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_terms_link(self, line, url):
        return self._handle_link(url)

    def handle_instance_organization_link(self, line, title, url):
        return self._handle_link(url, title)

    def _handle_link(self, url, title=None):
        if not url:
            content = "*Not configured, please contact Galaxy admin*"
            return (content, True)
        elif not title:
            title = url
        return (f"[{title}]({url})", True)

    def handle_invocation_time(self, line, invocation):
        content = literal_via_fence(invocation.create_time.strftime("%Y-%m-%d, %H:%M:%S"))
        return (content, True)

    def handle_dataset_name(self, line, hda):
        if hda.name:
            content = literal_via_fence(hda.name)
        else:
            content = "*No Dataset Name Available*"
        return (content, True)

    def handle_dataset_type(self, line, hda):
        if hda.ext:
            content = literal_via_fence(hda.ext)
        else:
            content = "*No Dataset Type Available*"
        return (content, True)

    def handle_error(self, container, line, error):
        return (line, False)


def to_basic_markdown(trans, internal_galaxy_markdown: str) -> str:
    """Replace Galaxy Markdown extensions with plain Markdown for PDF/HTML export."""
    directive_handler = ToBasicMarkdownDirectiveHandler(trans)
    plain_markdown = directive_handler.walk(trans, internal_galaxy_markdown)
    return plain_markdown


def to_html(basic_markdown: str) -> str:
    # Allow data: urls so we can embed images.
    html = sanitize_html(markdown.markdown(basic_markdown, extensions=["tables"]), allow_data_urls=True)
    return html


def to_pdf_raw(basic_markdown: str, css_paths: Optional[List[str]] = None) -> bytes:
    """Convert RAW markdown with specified CSS paths into bytes of a PDF."""
    css_paths = css_paths or []
    as_html = to_html(basic_markdown)
    directory = tempfile.mkdtemp("gxmarkdown")
    index = os.path.join(directory, "index.html")
    try:
        output_file = codecs.open(index, "w", encoding="utf-8", errors="xmlcharrefreplace")
        output_file.write(as_html)
        output_file.close()
        html = weasyprint.HTML(filename=index)
        stylesheets = [weasyprint.CSS(string=resource_string(__package__, "markdown_export_base.css"))]
        for css_path in css_paths:
            with open(css_path) as f:
                css_content = f.read()
            css = weasyprint.CSS(string=css_content)
            stylesheets.append(css)
        return html.write_pdf(stylesheets=stylesheets)
    finally:
        shutil.rmtree(directory)


def weasyprint_available() -> bool:
    return weasyprint is not None


def _check_can_convert_to_pdf_or_raise():
    """Checks if the HTML to PDF converter is available."""
    if not weasyprint_available():
        raise ServerNotConfiguredForRequest("PDF conversion service not available.")


def internal_galaxy_markdown_to_pdf(trans, internal_galaxy_markdown: str, document_type: PdfDocumentType) -> bytes:
    _check_can_convert_to_pdf_or_raise()
    basic_markdown = to_basic_markdown(trans, internal_galaxy_markdown)
    config = trans.app.config
    return to_branded_pdf(basic_markdown, document_type, config)


def generate_branded_pdf(
    request: GeneratePdfDownload, config: GalaxyAppConfiguration, short_term_storage_monitor: ShortTermStorageMonitor
):
    with storage_context(request.short_term_storage_request_id, short_term_storage_monitor) as target:
        raw_contents = to_branded_pdf(
            request.basic_markdown,
            request.document_type,
            config,
        )
        with open(target.path, "wb") as f:
            f.write(raw_contents)


def to_branded_pdf(basic_markdown: str, document_type: PdfDocumentType, config: GalaxyAppConfiguration) -> bytes:
    document_type_prologue = getattr(config, f"markdown_export_prologue_{document_type}s", "") or ""
    document_type_epilogue = getattr(config, f"markdown_export_epilogue_{document_type}s", "") or ""
    general_prologue = config.markdown_export_prologue or ""
    general_epilogue = config.markdown_export_epilogue or ""
    effective_prologue = document_type_prologue or general_prologue
    effective_epilogue = document_type_epilogue or general_epilogue
    branded_markdown = effective_prologue + basic_markdown + effective_epilogue
    css_paths = []
    general_css_path = config.markdown_export_css
    document_type_css_path = getattr(config, f"markdown_export_css_{document_type}s", None)
    if general_css_path and os.path.exists(general_css_path):
        css_paths.append(general_css_path)
    if document_type_css_path and os.path.exists(document_type_css_path):
        css_paths.append(document_type_css_path)
    return to_pdf_raw(branded_markdown, css_paths=css_paths)


def resolve_invocation_markdown(trans, invocation, workflow_markdown):
    """Resolve invocation objects to convert markdown to 'internal' representation.

    Replace references to abstract workflow parts with actual galaxy object IDs corresponding
    to the actual executed workflow. For instance:

        convert output=name -to- history_dataset_id=<id> | history_dataset_collection_id=<id>
        convert input=name -to- history_dataset_id=<id> | history_dataset_collection_id=<id>
        convert step=name -to- job_id=<id>

    Also expand/convert workflow invocation specific container sections into actual Galaxy
    markdown - these containers include: invocation_inputs, invocation_outputs, invocation_workflow.
    Hopefully this list will be expanded to include invocation_qc.
    """
    # TODO: convert step outputs?
    # convert step_output=index/name -to- history_dataset_id=<id> | history_dataset_collection_id=<id>

    def _section_remap(container, line):
        section_markdown = ""
        if container == "invocation_outputs":
            for output_assoc in invocation.output_associations:
                if not output_assoc.workflow_output.label:
                    continue

                if output_assoc.history_content_type == "dataset":
                    section_markdown += f"""#### Output Dataset: {output_assoc.workflow_output.label}
```galaxy
history_dataset_display(output="{output_assoc.workflow_output.label}")
```
"""
                else:
                    section_markdown += f"""#### Output Dataset Collection: {output_assoc.workflow_output.label}
```galaxy
history_dataset_collection_display(output="{output_assoc.workflow_output.label}")
```
"""
        elif container == "invocation_inputs":
            for input_assoc in invocation.input_associations:
                if not input_assoc.workflow_step.label:
                    continue

                if input_assoc.history_content_type == "dataset":
                    section_markdown += f"""#### Input Dataset: {input_assoc.workflow_step.label}
```galaxy
history_dataset_display(input="{input_assoc.workflow_step.label}")
```
"""
                else:
                    section_markdown += f"""#### Input Dataset Collection: {input_assoc.workflow_step.label}
```galaxy
history_dataset_collection_display(input={input_assoc.workflow_step.label})
```
"""
        else:
            return line, False
        return section_markdown, True

    def _remap(container, line):
        for workflow_instance_directive in ["workflow_display", "workflow_image"]:
            if container == workflow_instance_directive:
                stored_workflow_id = invocation.workflow.stored_workflow.id
                workflow_version = invocation.workflow.version
                return (
                    f"{workflow_instance_directive}(workflow_id={stored_workflow_id},workflow_checkpoint={workflow_version})\n",
                    False,
                )
        if container == "workflow_license":
            stored_workflow_id = invocation.workflow.stored_workflow.id
            return (
                f"workflow_license(workflow_id={stored_workflow_id})\n",
                False,
            )
        if container == "history_link":
            return (f"history_link(history_id={invocation.history.id})\n", False)
        if container == "invocation_time":
            return (f"invocation_time(invocation_id={invocation.id})\n", False)
        ref_object_type = None
        output_match = re.search(OUTPUT_LABEL_PATTERN, line)
        input_match = re.search(INPUT_LABEL_PATTERN, line)
        step_match = re.search(STEP_LABEL_PATTERN, line)

        def find_non_empty_group(match):
            for group in match.groups():
                if group:
                    return group

        target_match: Optional[Match]
        ref_object: Optional[Any]
        if output_match:
            target_match = output_match
            name = find_non_empty_group(target_match)
            ref_object = invocation.get_output_object(name)
        elif input_match:
            target_match = input_match
            name = find_non_empty_group(target_match)
            ref_object = invocation.get_input_object(name)
        elif step_match:
            target_match = step_match
            name = find_non_empty_group(target_match)
            invocation_step = invocation.step_invocation_for_label(name)
            if invocation_step and invocation_step.job:
                ref_object_type = "job"
                ref_object = invocation_step.job
            elif invocation_step and invocation_step.implicit_collection_jobs:
                ref_object_type = "implicit_collection_jobs"
                ref_object = invocation_step.implicit_collection_jobs
        else:
            target_match = None
            ref_object = None
        if ref_object:
            assert target_match  # tell type system, this is set when ref_object is set
            if ref_object_type is None:
                if ref_object.history_content_type == "dataset":
                    ref_object_type = "history_dataset"
                else:
                    ref_object_type = "history_dataset_collection"
            line = line.replace(target_match.group(), f"{ref_object_type}_id={ref_object.id}")
        return (line, False)

    workflow_markdown = _remap_galaxy_markdown_calls(
        _section_remap,
        workflow_markdown,
    )
    galaxy_markdown = _remap_galaxy_markdown_calls(_remap, workflow_markdown)
    return galaxy_markdown


def _remap_galaxy_markdown_containers(func, markdown):
    new_markdown = markdown

    searching_from = 0
    while True:
        from_markdown = new_markdown[searching_from:]
        match = re.search(GALAXY_FENCED_BLOCK, from_markdown)
        if match is not None:
            replace = match.group(1)
            (replacement, whole_block) = func(replace)
            if whole_block:
                start_pos = match.start()
                end_pos = match.end()
            else:
                start_pos = match.start(1)
                end_pos = match.end(1)
            start_pos = start_pos + searching_from
            end_pos = end_pos + searching_from
            new_markdown = new_markdown[:start_pos] + replacement + new_markdown[end_pos:]
            searching_from = start_pos + len(replacement)
        else:
            break

    return new_markdown


def _parse_directive_argument_value(arg_name: str, line: str) -> Optional[str]:
    arg_pattern = re.compile(rf"{arg_name}=\s*{ARG_VAL_CAPTURED_REGEX}\s*")
    match = re.search(arg_pattern, line)
    if not match:
        return None
    value = match.group(1)
    return value


def _remap_galaxy_markdown_calls(func, markdown):
    def _remap_container(container):
        matching_line = None
        for line in container.splitlines():
            if GALAXY_MARKDOWN_FUNCTION_CALL_LINE.match(line):
                assert matching_line is None
                matching_line = line

        if matching_line:
            match = GALAXY_MARKDOWN_FUNCTION_CALL_LINE.match(line)
            assert match  # already matched
            return func(match.group(1), f"{matching_line}\n")
        else:
            return (container, True)

    return _remap_galaxy_markdown_containers(_remap_container, markdown)


def _validate(*args, **kwds):
    """Light wrapper around validate_galaxy_markdown to throw galaxy exceptions instead of ValueError."""
    try:
        return validate_galaxy_markdown(*args, **kwds)
    except ValueError as e:
        raise MalformedContents(str(e))


__all__ = (
    "internal_galaxy_markdown_to_pdf",
    "ready_galaxy_markdown_for_export",
    "ready_galaxy_markdown_for_import",
    "resolve_invocation_markdown",
    "to_basic_markdown",
)



#!/usr/bin/env python

import optparse
import sqlite3
import sys
import tempfile

try:
    maketrans = str.maketrans
except AttributeError:
    from string import maketrans  # type: ignore[attr-defined,no-redef]


def stop_err(msg):
    sys.stderr.write(msg)
    sys.exit()


def solid2sanger(quality_string, min_qual=0):
    sanger = ""
    quality_string = quality_string.rstrip(" ")
    for qv in quality_string.split(" "):
        try:
            if int(qv) < 0:
                qv = "0"
            if int(qv) < min_qual:
                return False
            sanger += chr(int(qv) + 33)
        except Exception:
            pass
    return sanger


def Translator(frm="", to="", delete=""):
    if len(to) == 1:
        to = to * len(frm)
    trans = maketrans(frm, to)

    def callable(s):
        return s.translate(trans, delete)

    return callable


def merge_reads_qual(
    f_reads,
    f_qual,
    f_out,
    trim_name=False,
    out="fastq",
    double_encode=False,
    trim_first_base=False,
    pair_end_flag="",
    min_qual=0,
    table_name=None,
):
    # Reads from two files f_csfasta (reads) and f_qual (quality values) and produces output in three formats depending on out parameter,
    # which can have three values: fastq, txt, and db
    # fastq = fastq format
    # txt = space delimited format with defline, reads, and qvs
    # dp = dump data into sqlite3 db.
    # IMPORTNAT! If out = db two optins must be provided:
    #   1. f_out must be a db connection object initialized with sqlite3.connect()
    #   2. table_name must be provided
    if out == "db":
        cursor = f_out.cursor()
        sql = "create table %s (name varchar(50) not null, read blob, qv blob)" % table_name
        cursor.execute(sql)

    lines = []
    line = " "
    while line:
        for f in [f_reads, f_qual]:
            line = f.readline().rstrip("\n\r")
            while line.startswith("#"):
                line = f.readline().rstrip("\n\r")
            lines.append(line)

        if lines[0].startswith(">") and lines[1].startswith(">"):
            if lines[0] != lines[1]:
                stop_err(
                    "Files reads and quality score files are out of sync and likely corrupted. Please, check your input data"
                )

            defline = lines[0][1:]
            if trim_name and (defline[len(defline) - 3 :] == "_F3" or defline[len(defline) - 3 :] == "_R3"):
                defline = defline[: len(defline) - 3]

        elif not lines[0].startswith(">") and not lines[1].startswith(">") and len(lines[0]) > 0 and len(lines[1]) > 0:
            if trim_first_base:
                lines[0] = lines[0][1:]
            if double_encode:
                de = Translator(frm="0123.", to="ACGTN")
                lines[0] = de(lines[0])
            qual = solid2sanger(lines[1], int(min_qual))
            if qual:
                if out == "fastq":
                    f_out.write("@%s%s\n%s\n+\n%s\n" % (defline, pair_end_flag, lines[0], qual))
                if out == "txt":
                    f_out.write("%s %s %s\n" % (defline, lines[0], qual))
                if out == "db":
                    cursor.execute('insert into %s values("%s","%s","%s")' % (table_name, defline, lines[0], qual))
        lines = []


def main():
    usage = "%prog --fr F3.csfasta --fq R3.csfasta --fout fastq_output_file [option]"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option(
        "--fr", "--f_reads", metavar="F3_CSFASTA_FILE", dest="fr", help="Name of F3 file with color space reads"
    )
    parser.add_option(
        "--fq", "--f_qual", metavar="F3_QUAL_FILE", dest="fq", help="Name of F3 file with color quality values"
    )
    parser.add_option("--fout", "--f3_fastq_output", metavar="F3_OUTPUT", dest="fout", help="Name for F3 output file")
    parser.add_option(
        "--rr",
        "--r_reads",
        metavar="R3_CSFASTA_FILE",
        dest="rr",
        default=False,
        help="Name of R3 file with color space reads",
    )
    parser.add_option(
        "--rq",
        "--r_qual",
        metavar="R3_QUAL_FILE",
        dest="rq",
        default=False,
        help="Name of R3 file with color quality values",
    )
    parser.add_option("--rout", metavar="R3_OUTPUT", dest="rout", help="Name for F3 output file")
    parser.add_option(
        "-q",
        "--min_qual",
        dest="min_qual",
        default="-1000",
        help="Minimum quality threshold for printing reads. If a read contains a single call with QV lower than this value, it will not be reported. Default is -1000",
    )
    parser.add_option(
        "-t",
        "--trim_name",
        dest="trim_name",
        action="store_true",
        default=False,
        help="Trim _R3 and _F3 off read names. Default is False",
    )
    parser.add_option(
        "-f",
        "--trim_first_base",
        dest="trim_first_base",
        action="store_true",
        default=False,
        help="Remove the first base of reads in color-space. Default is False",
    )
    parser.add_option(
        "-d",
        "--double_encode",
        dest="de",
        action="store_true",
        default=False,
        help="Double encode color calls as nucleotides: 0123. becomes ACGTN. Default is False",
    )

    options, args = parser.parse_args()

    if not (options.fout and options.fr and options.fq):
        parser.error(
            """
        One or more of the three required paremetrs is missing:
        (1) --fr F3.csfasta file
        (2) --fq F3.qual file
        (3) --fout name of output file
        Use --help for more info
        """
        )

    fr = open(options.fr)
    fq = open(options.fq)
    f_out = open(options.fout, "w")

    if options.rr and options.rq:
        rr = open(options.rr)
        rq = open(options.rq)
        if not options.rout:
            parser.error("Provide the name for f3 output using --rout option. Use --help for more info")
        r_out = open(options.rout, "w")

        db = tempfile.NamedTemporaryFile()

        try:
            con = sqlite3.connect(db.name)
            cur = con.cursor()
        except Exception:
            stop_err("Cannot connect to %s\n") % db.name

        merge_reads_qual(
            fr,
            fq,
            con,
            trim_name=options.trim_name,
            out="db",
            double_encode=options.de,
            trim_first_base=options.trim_first_base,
            min_qual=options.min_qual,
            table_name="f3",
        )
        merge_reads_qual(
            rr,
            rq,
            con,
            trim_name=options.trim_name,
            out="db",
            double_encode=options.de,
            trim_first_base=options.trim_first_base,
            min_qual=options.min_qual,
            table_name="r3",
        )
        cur.execute("create index f3_name on f3( name )")
        cur.execute("create index r3_name on r3( name )")

        cur.execute("select * from f3,r3 where f3.name = r3.name")
        for item in cur:
            f_out.write("@%s%s\n%s\n+\n%s\n" % (item[0], "/1", item[1], item[2]))
            r_out.write("@%s%s\n%s\n+\n%s\n" % (item[3], "/2", item[4], item[5]))

    else:
        merge_reads_qual(
            fr,
            fq,
            f_out,
            trim_name=options.trim_name,
            out="fastq",
            double_encode=options.de,
            trim_first_base=options.trim_first_base,
            min_qual=options.min_qual,
        )

    f_out.close()


if __name__ == "__main__":
    main()



from typing import (
    List,
    Optional,
)

from pydantic import BaseModel
from sqlalchemy import (
    and_,
    select,
)
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm.scoping import scoped_session

from galaxy.exceptions import ObjectNotFound
from galaxy.model import StoreExportAssociation
from galaxy.model.base import transaction
from galaxy.schema.schema import ExportObjectType
from galaxy.structured_app import MinimalManagerApp


class StoreExportTracker:
    """
    Manages export-related metadata association of different types of exportable objects
    to remote file sources or repositories.
    """

    def __init__(self, app: MinimalManagerApp):
        self.app = app

    @property
    def session(self) -> scoped_session:
        return self.app.model.context

    def create_export_association(self, object_id: int, object_type: ExportObjectType) -> StoreExportAssociation:
        export_association = StoreExportAssociation(object_id=object_id, object_type=object_type)
        self.session.add(export_association)
        with transaction(self.session):
            self.session.commit()
        return export_association

    def set_export_association_metadata(self, export_association_id: int, export_metadata: BaseModel):
        try:
            stmt = select(StoreExportAssociation).where(StoreExportAssociation.id == export_association_id)
            export_association: StoreExportAssociation = self.session.execute(stmt).scalars().one()
        except NoResultFound:
            raise ObjectNotFound("Cannot set export metadata. Reason: Export association not found")
        export_association.export_metadata = export_metadata.model_dump_json()  # type:ignore[assignment]
        with transaction(self.session):
            self.session.commit()

    def get_export_association(self, export_association_id: int) -> StoreExportAssociation:
        try:
            stmt = select(StoreExportAssociation).where(StoreExportAssociation.id == export_association_id)
            export_association: StoreExportAssociation = self.session.execute(stmt).scalars().one()
        except NoResultFound:
            raise ObjectNotFound("Cannot get export association. Reason: Export association not found")
        return export_association

    def get_object_exports(
        self, object_id: int, object_type: ExportObjectType, limit: Optional[int] = None, offset: Optional[int] = None
    ) -> List[StoreExportAssociation]:
        stmt = (
            select(
                StoreExportAssociation,
            )
            .where(
                and_(
                    StoreExportAssociation.object_type == object_type,
                    StoreExportAssociation.object_id == object_id,
                    StoreExportAssociation.task_uuid.is_not(None),
                )
            )
            .order_by(StoreExportAssociation.create_time.desc())
        )
        if offset:
            stmt = stmt.offset(offset)
        if limit:
            stmt = stmt.limit(limit)
        return self.session.execute(stmt).scalars()  # type:ignore[return-value]



#!/usr/bin/env python
"""
Export a history to an archive file using attribute files.

usage: %prog history_attrs dataset_attrs job_attrs out_file
    -G, --gzip: gzip archive file
"""

import json
import optparse
import os
import shutil
import sys

from galaxy.files import ConfiguredFileSources
from galaxy.model.store import tar_export_directory
from galaxy.util import unicodify


def create_archive(export_directory, out_file, gzip=False):
    """Create archive from the given attribute/metadata files and save it to out_file."""
    try:
        tar_export_directory(export_directory, out_file, gzip)
        # Status.
        print("Created history archive.")
        return 0
    except Exception as e:
        print(f"Error creating history archive: {unicodify(e)}", file=sys.stderr)
        return 1
    finally:
        shutil.rmtree(export_directory, ignore_errors=True)


def main(argv=None):
    # Parse command line.
    parser = optparse.OptionParser()
    parser.add_option("-G", "--gzip", dest="gzip", action="store_true", help="Compress archive using gzip.")
    parser.add_option(
        "--galaxy-version", dest="galaxy_version", help="Galaxy version that initiated the command.", default=None
    )
    parser.add_option("--file-sources", type=str, help="file sources json")
    (options, args) = parser.parse_args(argv)

    gzip = bool(options.gzip)
    assert len(args) >= 2
    temp_directory = args[0]
    out_arg = args[1]

    destination_uri = None
    if "://" in out_arg:
        # writing to a file source instead of a dataset path.
        destination_uri = out_arg
        out_file = "./temp_out_archive"
    else:
        out_file = out_arg
    # Create archive.
    exit = create_archive(temp_directory, out_file, gzip=gzip)
    if destination_uri is not None and exit == 0:
        _write_to_destination(options.file_sources, os.path.abspath(out_file), destination_uri)
    return exit


def _write_to_destination(file_sources_path: str, out_file: str, destination_uri: str):
    file_sources = get_file_sources(file_sources_path)
    file_source_path = file_sources.get_file_source_path(destination_uri)
    file_source = file_source_path.file_source
    assert os.path.exists(out_file)
    file_source.write_from(file_source_path.path, out_file)


def get_file_sources(file_sources_path: str) -> ConfiguredFileSources:
    assert os.path.exists(file_sources_path), f"file sources path [{file_sources_path}] does not exist"
    with open(file_sources_path) as f:
        file_sources_as_dict = json.load(f)
    file_sources = ConfiguredFileSources.from_dict(file_sources_as_dict)
    return file_sources


if __name__ == "__main__":
    main()
